#include <iostream>
#include <sstream>

#include "typenameof.hh"
#include "stmt.hh"

stmt::stmt()
{
}
stmt::~stmt()
{
}

void stmt::Compile()
{
    std::cout << "    /* Compilation of "
              << typenameof(*this)
              << " not supported */\n";
}

static bool IsEntirelyIntegerCase(exprp exp)
{
    if(!exp) return false;
    
    if(dynamic_cast<expr_caseelse*>(&*exp))
    {
        return true;
    }
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(&*exp))
    {
        return IsEntirelyIntegerCase(e->a) && IsEntirelyIntegerCase(e->b);
    }
    if(expr_caseis* e = dynamic_cast<expr_caseis*>(&*exp))
    {
        if(e->token != (QBtoken)'=') return false;
        return IsEntirelyIntegerCase(e->ref);
    }
    if(expr_caserange* e = dynamic_cast<expr_caserange*>(&*exp))
    {
        return IsEntirelyIntegerCase(e->min) && IsEntirelyIntegerCase(e->max);
    }
    
    if(!exp->is_const()) return false;
    
    if(expr_number* e = dynamic_cast<expr_number*>(&*exp))
        if(e->is_int()) return true;
    
    TypeP type = exp->get_type();
    if(!type) return false;
    return
      (type->Type == TypeInfo::Integer
    || type->Type == TypeInfo::Long);
}

void stmt_print::OptimizeTail()
{
    std::string buffered_string;
    while(!params.empty())
    {
        std::pair<QBtoken,exprp>& last = params.back();
        if(last.first != (QBtoken)0) break;
        
        if(expr_string* str = dynamic_cast<expr_string*>(&*last.second))
        {
            buffered_string.insert(0, str->get());
            params.erase(--params.end());
        }
        else if(expr_number* num = dynamic_cast<expr_number*>(&*last.second))
        {
            if(!num->is_int()) break;
            std::stringstream tmp; tmp << num->get_i();
            buffered_string.insert(0, tmp.str());
            params.erase(--params.end());
        }
        else break;
    }
    if(!buffered_string.empty())
        params.push_back(std::make_pair((QBtoken)0, new expr_string(buffered_string)));
}

bool stmt_select::IsEntirelyIntegerCase() const
{
    for(std::list<std::pair<exprp, stmtp> >::const_iterator
        i = cases.begin(); i != cases.end(); ++i)
    {
        if(!::IsEntirelyIntegerCase(i->first)) return false;
    }
    return true;
}

void stmt_if::ForEachSubStmt(stmt_visitor& func)
{
    func.stmt_visit(thenp);
    for(std::list<
        std::pair<exprp, stmtp>
             >::const_iterator i = elseifp.begin(); i != elseifp.end(); ++i)
        func.stmt_visit(i->second);
    func.stmt_visit(elsep);
    stmt::ForEachSubStmt(func);
}
void stmt_if::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(condition);
    for(std::list<
        std::pair<exprp, stmtp>
             >::const_iterator i = elseifp.begin(); i != elseifp.end(); ++i)
        func.expr_visit(i->first);
    stmt::ForEachSubExpr(func);
}

void stmt_select::ForEachSubStmt(stmt_visitor& func)
{
    for(std::list<
        std::pair<exprp, stmtp>
             >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        func.stmt_visit(i->second);
    stmt::ForEachSubStmt(func);
}
void stmt_select::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(ref);
    for(std::list<
        std::pair<exprp, stmtp>
             >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        func.expr_visit(i->first);
    stmt::ForEachSubExpr(func);
}

void stmt_doloop::ForEachSubStmt(stmt_visitor& func)
{
    func.stmt_visit(bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_doloop::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(precondition);
    func.expr_visit(postcondition);
    stmt::ForEachSubExpr(func);
}

void stmt_for::ForEachSubStmt(stmt_visitor& func)
{
    func.stmt_visit(bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_for::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(var, true);
    func.expr_visit(begin);
    func.expr_visit(end);
    func.expr_visit(step);
    stmt::ForEachSubExpr(func);
}


void stmt_labelref::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_labelref::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(label);
    stmt::ForEachSubExpr(func);
}

void stmt_read::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_read::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target, true);
    stmt::ForEachSubExpr(func);
}


void stmt_let::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_let::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target, true);
    func.expr_visit(value);
    stmt::ForEachSubExpr(func);
}

void stmt_print::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_print::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target);
    func.expr_visit(using_string);
    for(std::list<std::pair<QBtoken, exprp> >::const_iterator
        i = params.begin(); i != params.end(); ++i)
        func.expr_visit(i->second);
    stmt::ForEachSubExpr(func);
}

void stmt_input::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_input::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(source);
    func.expr_visit(prompt);
    for(std::list<exprp>::const_iterator
        i = params.begin(); i != params.end(); ++i)
        func.expr_visit(*i, true);
    stmt::ForEachSubExpr(func);
}


void stmt_fileopen::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_fileopen::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(filename);
    func.expr_visit(mode_str);
    func.expr_visit(reclen);
    stmt::ForEachSubExpr(func);
}

void stmt_redim::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_redim::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target, true);
    for(unsigned a=0; a<dimensions.size(); ++a)
    {
        func.expr_visit(dimensions[a].first);
        func.expr_visit(dimensions[a].second);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_erase::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_erase::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target, true);
    stmt::ForEachSubExpr(func);
}

void stmt_subcall::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_subcall::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<params.size(); ++a)
    {
        func.set_func_call(name, a);
        func.expr_visit(params[a], maybe);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_label::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_label::ForEachSubExpr(expr_visitor& func)
{
    stmt::ForEachSubExpr(func);
}


void stmt_mid_s::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_mid_s::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(target, true);
    func.expr_visit(start);
    func.expr_visit(length);
    func.expr_visit(value);
    stmt::ForEachSubExpr(func);
}


void stmt_lib_stmt::ForEachSubStmt(stmt_visitor& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_lib_stmt::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<params.size(); ++a)
        func.expr_visit(params[a]);
    stmt::ForEachSubExpr(func);
}
